'''Test wal parsers'''
# pylint: disable=C0103, W0611, C0116, W0201
import unittest

from lark import Lark

from wal.reader import read, WAL_GRAMMAR, ParseError, read_wal_sexpr, read_wal_sexprs
from wal.ast_defs import Symbol as S, WList
from wal.ast_defs import Operator as Op

test_readers = {}
def read_test(code, start):
    if start not in test_readers:
        test_readers[start] = Lark(WAL_GRAMMAR, start=start, parser='lalr', propagate_positions=True)

    return read(code, test_readers[start])


class BasicParserTest(unittest.TestCase):
    '''Test basic parsers'''

    def test_boolean(self):
        '''Test boolean parser'''
        self.assertEqual(read_wal_sexpr('#t'), 1)
        self.assertEqual(read_wal_sexpr('#f'), 0)

    def test_binary(self):
        '''Test bin int parser'''
        reader = lambda c: read_test(c, 'bin_int')
        self.assertEqual(reader('0b0'), 0)
        self.assertEqual(reader('0b1'), 1)
        self.assertEqual(reader('0b11'), 3)
        self.assertEqual(reader('0b101010110101'), 2741)
        self.assertRaises(ParseError, reader, '4')
        self.assertRaises(ParseError, reader, 'a')
        self.assertRaises(ParseError, reader, 'u')
        self.assertRaises(ParseError, reader, '(')

    def test_decimal_int(self):
        '''Test dec int parser'''
        reader = lambda c: read_test(c, 'int')
        self.assertEqual(reader('4'), 4)
        self.assertEqual(reader('46'), 46)
        self.assertEqual(reader('-4'), -4)
        self.assertEqual(reader('-46'), -46)
        self.assertRaises(ParseError, reader, '4 6')
        self.assertRaises(ParseError, reader, 'a')
        self.assertRaises(ParseError, reader, '- 4')

    def test_hexadecimal_int(self):
        '''Test hex int parser'''
        reader = lambda c: read_test(c, 'hex_int')
        self.assertEqual(reader('0x4'), 4)
        self.assertEqual(reader('0x46'), 70)
        self.assertEqual(reader('0x4ffaf6'), 0x4ffaf6)
        self.assertRaises(ParseError, reader, '4 6')
        self.assertRaises(ParseError, reader, 'a')
        self.assertRaises(ParseError, reader, '0xh')

    def test_float(self):
        '''Test parsing floats'''
        reader_float = lambda c: read_test(c, 'float')
        reader_sexpr = lambda c: read_test(c, 'sexpr')
        pass_cases = [x / 3.2 for x in range(-500, 500, 15)]
        for case in pass_cases:
            self.assertEqual(reader_float(str(case)), case)
            self.assertEqual(reader_sexpr(str(case)), case)

    def test_text(self):
        '''Test text parser'''
        reader = lambda c: read_test(c, 'string')
        self.assertEqual(reader('"foo"'), 'foo')
        self.assertEqual(reader('"1234"'), '1234')
        self.assertEqual(reader('"\'bar\'"'), '\'bar\'')
        self.assertRaises(ParseError, reader, '1234')

    def test_base_symbol(self):
        '''Test symbol parser'''
        reader_base = lambda c: read_test(c, 'base_symbol')
        reader_symbol = lambda c: read_test(c, 'symbol')
        pass_cases = ['valid', '_valid', 'Valid', 'v123lid', 'va$lid', 'state::bla=>x', 'x2']
        for case in pass_cases:
            self.assertEqual(reader_base(case), S(case))
            self.assertEqual(reader_symbol(case), S(case))

        fail_cases = ['1valid', '%valid%', '\t', '1list', '-d21']
        for case in fail_cases:
            self.assertRaises(ParseError, reader_base, case)
            self.assertRaises(ParseError, reader_symbol, case)

    def test_scoped_symbol(self):
        '''Test symbol parser'''
        reader = lambda c: read_test(c, 'scoped_symbol')
        pass_cases = ['~valid', '~_valid', '~Valid', '~v123li/d', '~va$lid']
        for case in pass_cases:
            self.assertEqual(reader(case), WList([Op.RESOLVE_SCOPE, S(case[1:])]))

        fail_cases = ['1valid', ' valid',
                      'valid%', '  ', '\t', '1list', '-d21']
        for case in fail_cases:
            self.assertRaises(ParseError, reader, case)

    def test_grouped_symbol(self):
        '''Test symbol parser'''
        reader = lambda c: read_test(c, 'grouped_symbol')
        pass_cases = ['#valid', '#_valid', '#Valid', '#v123li/d', '#va$lid']
        for case in pass_cases:
            self.assertEqual(reader(case), WList([Op.RESOLVE_GROUP, S(case[1:])]))

        fail_cases = ['1valid', ' valid',
                      'valid%', '  ', '\t', '1list', '-d21']
        for case in fail_cases:
            self.assertRaises(ParseError, reader, case)

    def test_operators(self):
        '''Test built-in operators'''
        reader = lambda c: read_test(c, 'atom')
        for builtin in Op:
            self.assertEqual(reader(builtin.value),
                             Op(builtin.value))


class SexprParserTest(unittest.TestCase):
    '''Test all parsers related to s-expressions'''

    def test_sexpr(self):
        '''Test simple s-expressions'''
        reader = lambda c: read_test(c, 'sexpr')
        self.assertEqual(reader('5'), 5)
        self.assertEqual(reader('"foo"'), 'foo')
        self.assertEqual(reader('valid'), S('valid'))

        self.assertEqual(reader('(+ 1 2)'), [Op.ADD, 1, 2])
        # self.assertEqual(reader('(+ 1 (- 3 4))'),
        #                  [Op.ADD, 1, [Op.SUB, 3, 4]])

    def test_quoting(self):
        '''Test qouted sexprs'''
        reader = lambda c: read_test(c, 'sexpr')
        self.assertEqual(reader('\'5'), [Op.QUOTE, 5])
        self.assertEqual(reader('\'valid'),
                         [Op.QUOTE, S('valid')])

    def test_sexpr_whtitespace(self):
        '''Test parsing of simple statements'''
        reader = lambda c: read_test(c, 'sexpr')
        golden = [Op.ADD, 1, 2]
        self.assertEqual(reader('(+ 1 2)'), golden)
        self.assertEqual(reader(' (+ 1 2)'), golden)
        self.assertEqual(reader('\n (+ 1 2) '), golden)
        self.assertEqual(reader('\n (+ 1 2) \n '), golden)

    def test_sexpr_simple_nested(self):
        reader = lambda c: read_test(c, 'sexpr')
        golden = [Op.SUB, [Op.ADD, 1, 2]]
        self.assertEqual(reader('(- (+ 1 2))'), golden)

    def test_sexpr_quoted(self):
        reader = lambda c: read_test(c, 'sexpr')
        golden = [Op.QUOTE, [Op.ADD, 1, 2]]
        self.assertEqual(reader("'(+ 1 2)"), golden)

    def test_timed_sexpr(self):
        '''Test relative expression evaluation'''
        reader = lambda c: read_test(c, 'sexpr')
        self.assertEqual(reader('valid@5'), [Op.REL_EVAL, S('valid'), 5])
        self.assertEqual(reader('valid@-5'), [Op.REL_EVAL, S('valid'), -5])
        self.assertEqual(reader('valid@(+ 1 2)'), [Op.REL_EVAL, S('valid'), [Op.ADD, 1, 2]])
        # self.assertEqual(reader('(+ valid 2)@5'), [Op.REL_EVAL, [Op.ADD, S('valid'), 2], 5])
        # self.assertEqual(reader('(+ valid 2)@x'), [Op.REL_EVAL, [Op.ADD, S('valid'), 2], S('x')])
        # self.assertEqual(reader('(+ valid 2)@(* x 2)'), [Op.REL_EVAL, [Op.ADD, S('valid'), 2], [Op.MUL, S('x'), 2]])
        self.assertRaises(ParseError, reader, 'valid @5')
        self.assertRaises(ParseError, reader, 'valid@ 5')

        # self.assertEqual(reader('(+ x y)@-1'), [Op.REL_EVAL, [Op.ADD, S('x'), S('y')], -1])


class SimpleProgramTest(unittest.TestCase):
    '''Test parse simple programs'''

    def test_programs(self):
        '''Test a simple program with two statements'''
        # p = ''''''
        # golden = []
        # self.assertEqual(read_wal_sexprs(p), golden)

        # p = '''

        # '''
        # golden = []
        # self.assertEqual(read_wal_sexprs(p), golden)

        p = '''(+ 1 2)'''
        golden = [[Op.ADD, 1, 2]]
        self.assertEqual(read_wal_sexprs(p), golden)

        p = '''( +   1 (- x 5))'''
        golden = [[Op.ADD, 1, [Op.SUB, S('x'), 5]]]
        self.assertEqual(read_wal_sexprs(p), golden)

        p = '''(+ 1 (- x 5)) x'''
        golden = [[Op.ADD, 1, [Op.SUB, S('x'), 5]], S('x')]
        #a = read_wal_sexprs(p)
        #print(a.pretty())
        self.assertEqual(read_wal_sexprs(p), golden)


        p = '''
        (set [x 5])

        (print (+ 1 (- x 5)))'''
        golden = [[Op.SET, [S('x'), 5]], [Op.PRINT, [Op.ADD, 1, [Op.SUB, S('x'), 5]]]]
        self.assertEqual(read_wal_sexprs(p), golden)

        p = '''
        (set [x 5]) ;comment
        ; comment
           ;comment
        (print (+ 1 (- x 5)))'''
        self.assertEqual(read_wal_sexprs(p), golden)

        p = ''';(+ 1 2)'''
        golden = []
        self.assertEqual(read_wal_sexprs(p), golden)


class SimpleProgramFileTest(unittest.TestCase):
    '''Test parse simple programs from files'''

    def test_program1(self):
        '''Test a simple program with two statements'''
        with open('tests/files/p1.wal', encoding='utf-8') as f:
            p = f.read()
            golden = [[Op.PRINT, 'hello, test'], [Op.ADD, S('x'), 2]]
            self.assertEqual(read_wal_sexprs(p), golden)

    def test_program2(self):
        '''Test a simple program with a when condition'''
        with open('tests/files/p2.wal', encoding='utf-8') as f:
            p = f.read()
            golden = [[S('when'), [Op.AND, S('i_valid'), S('i_ready')], [Op.PRINT, S('i_data')]]]
            self.assertEqual(read_wal_sexprs(p), golden)

    def test_program3(self):
        '''Test a simple program with a syntax error'''
        with open('tests/files/p3.wal', encoding='utf-8') as f:
            p = f.read()
            self.assertRaises(ParseError, read_wal_sexprs, p)



'''Generic trace class '''
# pylint: disable=E1101, W0201

class Trace:
    '''A generic class for representing waveforms'''
    SCOPE_SEPERATOR = '^'
    SPECIAL_SIGNALS = ['SIGNALS', 'SIGNALS-NO-ALIAS', 'VIRTUAL-SIGNALS', 'LOCAL-SIGNALS', 'INDEX', 'MAX-INDEX', 'TS', 'TRACE-NAME', 'TRACE-FILE', 'SCOPES', 'LOCAL-SCOPES']
    SPECIAL_SIGNALS_SET = set(SPECIAL_SIGNALS)


    def __init__(self, tid, filename, container):
        self.tid = tid
        self.filename = filename
        self.container = container
        self.virtual_signals = {}
        self.index = 0


    def set(self, index=0):
        '''Set the index of this trace'''
        self.index = index


    def step(self, steps=1):
        '''Add steps to the index of this trace.
        If the resulting index is invalid the tid of this trace is returned.'''
        rel_index = self.index + steps

        if rel_index < 0 or rel_index > self.max_index:
            return self.tid

        self.index = rel_index
        return None


    def signal_value(self, name, offset, scope=''):
        '''Get the value of signal name at current time + offset.'''
        rel_index = self.index + offset

        res = None
        # handle special variables
        if 0 <= rel_index <= self.max_index:
            if name in Trace.SPECIAL_SIGNALS:
                if name == 'SIGNALS':
                    res = self.get_all_signals()
                    if len(self.container.traces) > 1:
                        res = [f'{self.tid}{Trace.SCOPE_SEPERATOR}{signal}' for signal in res]
                elif name == 'SIGNALS-NO-ALIAS':
                    res = self.get_all_signals_by_handle()
                    if len(self.container.traces) > 1:
                        res = [f'{self.tid}{Trace.SCOPE_SEPERATOR}{signal}' for signal in res]
                elif name == 'VIRTUAL-SIGNALS':
                    res = self.get_virtual_signals()
                    if len(self.container.traces) > 1:
                        res = [f'{self.tid}{Trace.SCOPE_SEPERATOR}{signal}' for signal in res]
                elif name == 'LOCAL-SIGNALS':
                    if scope == '':
                        res = [s for s in self.rawsignals if '.' not in s]
                    else:
                        def in_scope(signal):
                            prefix_ok = signal.startswith(scope + '.')
                            not_in_sub_scope = '.' not in signal[len(scope) + 1:]
                            return prefix_ok and not_in_sub_scope

                        res = list(filter(in_scope, self.signals))
                elif name == 'INDEX':
                    res = self.index
                elif name == 'MAX-INDEX':
                    res = self.max_index
                elif name == 'TS':
                    res = self.ts
                elif name == 'TRACE-NAME':
                    res = self.tid
                elif name == 'TRACE-FILE':
                    res = self.filename
                elif name == 'LOCAL-SCOPES':
                    if scope != '':
                        scope += '.'
                    res = [s for s in self.scopes if (s.startswith(scope)) and ('.' not in s[len(scope) + 1:])]
                elif name == 'SCOPES':
                    res = self.scopes
            elif name in self.virtual_signals:
                res = self.virtual_signals[name].value
            else:
                res = self.access_signal_data(name, rel_index)
        elif rel_index >= self.max_index:
            res = self.access_signal_data(name, self.max_index)
        else:
            raise ValueError(f'can not access {name} at negative timestamp')

        return res

    def signal_width(self, name):
        '''Returns the signal width'''
        raise NotImplementedError

    @property
    def ts(self):
        '''Converts the index to the current timestamp.'''
        return self.timestamps[self.index]

    def set_sampling_points(self, new_indices):
        '''Updates the indices at which data is sampled'''
        new_timestamps = [self.all_timestamps[i] for i in new_indices]
        self.timestamps = list(dict.fromkeys(new_timestamps))
        self.timestamps = dict(enumerate(self.timestamps))
        # stores current time stamp
        self.index = 0
        self.max_index = len(self.timestamps.keys()) - 1

    def add_virtual_signal(self, signal):
        '''Adds a virtual signal to this trace'''
        self.signals.add(signal.name)
        self.virtual_signals[signal.name] = signal

    def get_all_signals(self):
        return list(self.rawsignals) + self.get_virtual_signals()
    
    def get_virtual_signals(self):
        return list(self.virtual_signals.keys())

    def get_all_signals_by_handle(self):
        return list(self.rawsignals_by_handle)

    def set_max_index(self, new_max_index):
        self.max_index = min(new_max_index, self.max_index)
        return self.max_index



'''Implementations for WAL related functions such as loading and unloading traces.'''
import os

from wal.ast_defs import Operator, Symbol, WList
from wal.util import wal_decode
from wal.reader import read_wal_sexprs
from wal.passes import expand, optimize, resolve
from wal.repl import WalRepl


def op_load(seval, args):
    '''Load trace from file filename under id tid '''
    assert len(args) == 1 or len(args) == 2, 'load: expects two arguments (load filename:str|symbol tid:str|symbol?)'
    filename = seval.eval(args[0])
    tid = None
    if len(args) == 2:
      tid = seval.eval(args[1]) if len(args) == 2 else None
      assert isinstance(tid, (str, Symbol)), 'load: second argument must be str or symbol'
      tid = tid.name if isinstance(tid, Symbol) else tid

    assert isinstance(filename, (str, Symbol)), 'load: first argument must be str or symbol'
    filename = filename if isinstance(filename, str) else filename.name
    seval.traces.load(filename, tid)


def op_unload(seval, args):
    '''Removes the trace with id t from WAL '''
    assert len(args) == 1, 'unload: expects one argument (unload trace:symbol)'
    tid = seval.eval(args[0])
    assert isinstance(tid, (str, Symbol)), 'unload: argument must be str or symbol'
    tid = tid if isinstance(tid, str) else tid.name
    return seval.traces.unload(tid)


def op_step(seval, args):
    '''Steps the trace or all traces by offset'''
    assert seval.traces.traces, 'step: no traces loaded'

    def do_step(tid, steps=1):
        if isinstance(tid, str):
            return seval.traces.step(tid=tid, steps=steps)
        if isinstance(tid, Symbol):
            return seval.traces.step(tid=tid.name, steps=steps)

        raise ValueError('step: arguments must be either str or symbol')

    # step all traces +1
    if len(args) == 0:
        res = seval.traces.step()
    elif len(args) == 1:
        evaluated = seval.eval(args[0])
        if isinstance(evaluated, int):
            res = seval.traces.step(evaluated)
        else:
            res = do_step(args[0])
    else:
        evaluated = seval.eval(args[-1])
        assert isinstance(evaluated, int), 'step: last argument must be int'
        steps = evaluated
        res = []
        for tid in args[:-1]:
            step = do_step(tid, steps)
            if step:
                res.append(step)

    return res == []


def op_eval_file(seval, args):
    '''Executes WAL module and adds definitions to WAL context'''
    assert len(args) > 0, 'require: expects at least one argument (require module:str|symbol+)'
    assert all(isinstance(s, (str, Symbol)) for s in args), 'require: arguments must be symbols or strings'
    args = [arg.name if isinstance(arg, Symbol) else arg for arg in args]

    for module in args:

        def wal_file_exists(name):
            for path in seval.walpath:
                fname = f'{path}/{name}'
                if os.path.isfile(fname):
                    return fname

            return False

        sexprs = None
        if name := wal_file_exists(module + '.wo'):
            sexprs = wal_decode(name)
        elif name := wal_file_exists(module + '.wal'):
            with open(name, 'r', encoding='utf-8') as file:
                sexprs = read_wal_sexprs(file.read(), name)
        else:
            print(f'require: cant find file {module}.wal')
            raise FileNotFoundError

        if sexprs:
            for sexpr in sexprs:
                expanded = expand(seval, sexpr, parent=seval.global_environment)
                optimized = optimize(expanded)
                resolved = resolve(optimized, start=seval.global_environment.environment)
                seval.eval(resolved)


def op_require(seval, args):
    'Make bindings from a module available in current scope'
    for spec in args:
        if isinstance(spec, Symbol):
            # imports all bindings from the module
            module_name = spec.name
            module = seval.environment.read(module_name)
            for name, binding in module.items():
                assert name in module, f'require: can not require {name} from module {module_name}'
                seval.environment.define(name, binding)
        elif isinstance(spec, (WList, list)):
            if len(spec) > 1 and spec[0] == Symbol('rename-in'):
                # imports only the selected bindings and renames them
                module_name = spec[1].name
                module = seval.environment.read(module_name)
                for rename_binding in spec[2:]:
                    assert isinstance(rename_binding, (WList, list)) \
                        and len(rename_binding) == 2, 'require: rename-in arguments must be tuples'
                    old_name = rename_binding[0].name
                    new_name = rename_binding[1].name
                    assert old_name in module, f'require: can not require {old_name} from module {module_name}'
                    seval.environment.define(new_name, module[old_name])
            elif len(spec) == 3 and spec[0] == Symbol('only-in'):
                # imports only the selected bindings and renames them
                module_name = spec[1].name
                module = seval.environment.read(module_name)
                selected = spec[2]
                assert isinstance(selected, (WList, list)), 'require: only-in expects a list of bindings'
                for binding in selected:
                    assert isinstance(binding, Symbol), 'require: only-in bindings must be symbols'
                    name = binding.name
                    assert name in module, f'require: can not require {name} from module {module_name}'
                    seval.environment.define(name, module[name])
            elif len(spec) == 3 and spec[0] == Symbol('prefix-in'):
                # imports all bindings from the module and prefixes them
                prefix = spec[1]
                assert isinstance(prefix, Symbol), 'require: prefix-in prefix must be a symbol'
                module_name = spec[2].name
                module = seval.environment.read(module_name)
                for name, binding in module.items():
                    assert name in module, f'require: can not require {name} from module {module_name}'
                    seval.environment.define(prefix.name + name, binding)


def op_repl(seval, args):
    '''Starts a REPL in the current context'''
    try:
        WalRepl(seval.wal, intro=WalRepl.dyn_intro).cmdloop()
    except KeyboardInterrupt:
        pass


def op_is_signal(seval, args):
    assert len(args) == 1, 'signal?: expects exactly one argument (signal? s:symbol|string)'
    name = seval.eval(args[0])
    assert isinstance(name, (Symbol, str)), 'signal?: expects exactly one argument (signal? s:symbol|string)'
    return seval.traces.contains(name if isinstance(name, str) else name.name)


wal_operators = {
    Operator.LOAD.value: op_load,
    Operator.UNLOAD.value: op_unload,
    Operator.STEP.value: op_step,
    Operator.EVAL_FILE.value: op_eval_file,
    Operator.REQUIRE.value: op_require,
    Operator.REPL.value: op_repl,
    Operator.IS_SIGNAL.value: op_is_signal,
}



'''WAL Reader'''

import ast
from lark import Lark, Transformer
from lark import UnexpectedToken, UnexpectedEOF, UnexpectedCharacters
from lark.visitors import v_args
from lark.exceptions import VisitError
from wal.ast_defs import Symbol, Operator, Unquote, UnquoteSplice, operators, WList
from wal.ast_defs import Symbol as S

WAL_GRAMMAR = r"""
    _NL: /(\r?\n)+/
    _COMMENT: _WS? ";" /.*/
    _WS: WS
    _BASH_LINE : /\s*#![^\n]+\n/

    sexpr_strict: (atom | list | quoted | quasiquoted | unquote | unquote_splice)
    timed_sexpr : sexpr_strict "@" sexpr_strict
    _INTER : _COMMENT | _WS
    sexpr : _INTER* (sexpr_strict | timed_sexpr) _INTER*
    quoted : "'" sexpr
    quasiquoted : "`" sexpr
    unquote : "," sexpr
    unquote_splice : ",@" sexpr
    sexpr_list : _BASH_LINE? (_INTER* | sexpr*)

    atom : string
         | bool
         | operator
         | symbol
         | float
         | int

    int : dec_int | bin_int | hex_int
    float : /[+-]?[0-9]+\.[0-9]*/
    dec_int : /[+-]?[0-9]+/
    bin_int : /0b[0-1]+/
    hex_int : /0x[0-9a-fA-F]+/

    bool : true | false
    true : "true"
    false : "false"
    symbol : base_symbol | scoped_symbol | grouped_symbol | bit_symbol | sliced_symbol
    !operator: "+" | "-" | "*" | "/" | "&&" | "||" | "=" | "!=" | ">" | "<" | ">=" | "<=" | "!" | "**"

    scoped_symbol : "~" base_symbol
    grouped_symbol : "#" base_symbol
    !base_symbol : /[a-zA-Z_\.][=$\*\/>:\.\-_\?=%§^!\\~+<>|,\w]*/ | /\\\S+/
    bit_symbol : sexpr_strict "[" sexpr "]"
    sliced_symbol : sexpr_strict "[" sexpr ":" sexpr "]"

    string : ESCAPED_STRING

    list : "(" [sexpr*] ")"
           | "[" [sexpr*] "]"
           | "{" [sexpr*] "}"

    %import common.ESCAPED_STRING
    %import common.WS
    """


class TreeToWal(Transformer):

    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    def line_info(self, meta):
        return {
            'filename': self.filename,
            'line': meta.line,
            'end_line': meta.end_line,
            'column': meta.column,
            'end_column': meta.end_column
        }
    
    '''Transformer to create valid WAL expressions form parsed data'''
    string = lambda self, s: ast.literal_eval(s[0])

    @v_args(meta=True)
    def symbol(self, meta, sym):
        '''Converts symbols to operators if sym is a valid operator '''
        sym = sym[0]
        if isinstance(sym, Symbol):
            return Operator(sym.name) if sym.name in operators else sym

        return sym

    @v_args(meta=True)
    def quoted(self, meta, q):
        return WList([Operator.QUOTE, q[0]], line_info=self.line_info(meta))

    @v_args(meta=True)
    def quasiquoted(self, meta, q):
        return WList([Operator.QUASIQUOTE, q[0]], line_info=self.line_info(meta))

    unquote = lambda self, q: Unquote(q[0])
    unquote_splice = lambda self, q: UnquoteSplice(q[0])

    operator = lambda self, o: Operator(o[0].value)

    @v_args(meta=True)
    def simple_symbol(self, meta, s):
        sym = s[0]
        sym.line_info(meta)
        return sym

    @v_args(meta=True)
    def base_symbol(self, meta, s):
        return S(''.join(list(map(lambda x: x.value, s))), line_info=self.line_info(meta))

    @v_args(meta=True)
    def scoped_symbol(self, meta, s):
        return WList([Operator.RESOLVE_SCOPE, s[0]], line_info=self.line_info(meta))

    @v_args(meta=True)
    def grouped_symbol(self, meta, s):
        if s[0].name == 't':
            return True
        if s[0].name == 'f':
            return False

        return WList([Operator.RESOLVE_GROUP, s[0]], line_info=self.line_info(meta))

    @v_args(meta=True)
    def bit_symbol(self, meta, s):
        return WList([Operator.SLICE, s[0], s[1]], line_info=self.line_info(meta))

    @v_args(meta=True)
    def sliced_symbol(self, meta, s):
        return WList([Operator.SLICE, s[0], s[1], s[2]], line_info=self.line_info(meta))

    int = lambda self, i: i[0]
    float = lambda self, f: float(f[0])
    bin_int = lambda self, i: int(i[0], 2)
    dec_int = lambda self, i: int(i[0])
    hex_int = lambda self, i: int(i[0], 16)

    atom = lambda self, a: a[0]

    @v_args(meta=True)
    def list(self, meta, data):
        return WList(data, line_info=self.line_info(meta))

    
    bool = lambda self, b: b[0]
    true = lambda self, _: True
    false = lambda self, _: False

    @v_args(meta=True)
    def sexpr(self, meta, s):
        res = None
        if isinstance(s[0], list):
            res = WList(s[0], line_info=self.line_info(meta))
        elif s:
            res = s[0]

        return res

    @v_args(meta=True)
    def timed_sexpr(self, meta, s):
        return WList([Operator.REL_EVAL, s[0], s[1]], line_info=self.line_info(meta))

    @v_args(meta=True)
    def sexpr_list(self, meta, data):
        return WList(data, line_info=self.line_info(meta))

    @v_args(meta=True)
    def sexpr_strict(self, meta, s):
        return s[0] if s else None


def read(code, reader, filename=''):
    try:
        parsed = reader.parse(code)
        return TreeToWal(filename).transform(parsed)
    except UnexpectedEOF as u:
        context = u.get_context(code)
        msg = f'Unexpected end of file at  at line {u.line}:{u.column}.\nDid you forget a closing )?'
        raise ParseError(context, msg) from u
    except UnexpectedToken as u:
        context = u.get_context(code)
        msg = f'Unexpected "{u.token}" at line {u.line}:{u.column}'
        raise ParseError(context, msg) from u
    except UnexpectedCharacters as u:
        context = u.get_context(code)
        msg = f'Unexpected "{u.char}" at line {u.line}:{u.column}'
        raise ParseError(context, msg) from u
    except VisitError as u:
        context = u.__context__
        raise ParseError("", str(context)) from u


reader_sexpr = Lark(WAL_GRAMMAR, start='sexpr', parser='lalr', propagate_positions=True)
def read_wal_sexpr(code, filename=''):
    return read(code, reader_sexpr, filename=filename)


reader_sexprs = Lark(WAL_GRAMMAR, start='sexpr_list', parser='lalr', propagate_positions=True)
def read_wal_sexprs(code, filename=''):
    sexprs = read(code, reader_sexprs, filename=filename)
    return sexprs


class ParseError(Exception):
    '''Exception thrown when a syntax error is detected'''

    def __init__(self, context, message):
        super().__init__(context + '\n' + message)

        self.context = context
        self.message = message

    def show(self):
        print(self.context)
        print(self.message)



'''Main wawk command line tool entry point'''
import argparse
import os
import sys

from wal.core import Wal
from wal.util import wal_str
from .parser import parse_wawk
from .ast_defs import AST

__version__ = '0.2'

class Arguments:  # pylint: disable=too-few-public-methods
    '''Wrapper class for argument parser'''

    def __init__(self):
        parser = argparse.ArgumentParser()
        self.parser = parser

        parser.add_argument('program_path', help='path to WAWK file')
        parser.add_argument('trace', help='path to trace file')
        parser.add_argument('args', nargs='*',
                            default=None, help='runtime arguments')
        parser.add_argument('-v', '--version', action='version',
                            version=f'%(prog)s {__version__}')
        parser.add_argument('-o', '--output', nargs='?', default=None, help='Do not run but only write WAL program to file.')

    def parse(self):
        '''Parse program arguments and check minimal requirements'''
        args = self.parser.parse_args()

        return args


def run():
    '''Entry point for the application script'''

    arg_parser = Arguments()
    args = arg_parser.parse()
    sys.setrecursionlimit(5000)

    try:
        with open(args.program_path, 'r', encoding='UTF-8') as program_file:
            ast = AST(parse_wawk(program_file.read()), args.trace)
    except FileNotFoundError as exception:
        print(exception)
        return os.EX_NOINPUT
    except IOError as exception:
        print(exception)
        return os.EX_IOERR

    wal = Wal()
    wal.eval_context.global_environment.define('args', args.args)

    if args.output:
        with open(args.output, 'w') as f:
            stmts, symbols = ast.emit()
            for stmt in stmts:
                f.write(wal_str(stmt) + '\n\n')
    else:
        exprs, symbols = ast.emit()
        wal.load(args.trace, 'WAWK_TRACE', keep_signals=symbols)
        for expr in exprs:
            wal.eval(expr)

    return os.EX_OK


if __name__ == "__main__":
    sys.exit(run())



'''Trace implementation for the FST file format '''

from wal.trace.trace import Trace

class TraceVirtual(Trace):
    '''Holds data for one virtual trace.'''

    def __init__(self, tid, max_index, container):
        super().__init__(tid, 'virtual', container)
        self.signals = set()
        self.rawsignals = self.signals
        self.max_index = max_index
        self.timestamps = range(max_index)
        self.scopes = []

    def dump_vcd(self):
        '''Dumps this trace in vcd format'''
         # pylint: disable=C0103
        with open(self.tid + '.vcd', 'w', encoding='utf-8') as f:
            f.write('$version\n    WAL\n$end\n')
            f.write('$timescale\n    1ps\n$end\n')
            f.write(f'$scope module {self.tid} $end\n')
            for signal in self.signals:
                f.write(f'$var reg 32 {signal} {signal} [31:0] $end\n')

            f.write('$upscope $end\n')
            f.write('$enddefinitions $end\n')

            old_index = self.index
            last_values = {}
            for index in range(int(self.max_index/2)):
                self.index = index
                f.write(f'#{index}\n')
                if index == 0:
                    f.write('$dumpvars\n')

                for signal in self.signals:
                    value = self.signal_value(signal, index)
                    if (signal not in last_values) or (value != last_values[signal]):
                        width = 32
                        last_values[signal] = value
                        txt = f'{value:0{width}b}'.format(value=value, width=width)
                        f.write(f'b{txt} {signal}\n')

            self.index = old_index

    def signal_width(self, name):
        return 32

    def get_all_signals(self):
        return list(self.virtual_signals.keys())
    
    def get_all_signals_by_handle(self):
        return self.get_all_signals()


'''Setup file for the wal-lang package'''
import pathlib
from setuptools import setup, find_packages

with open('wal/version.py', encoding="utf8") as f:
    exec(f.read())

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='wal-lang',
    version=__version__, # noqa: F821
    description='Wal - Wavefile Analysis Language',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://wal-lang.org',
    author='Lucas Klemmer',
    author_email='support@wal-lang.org',
    classifiers=[
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: OS Independent'
    ],
    keywords=['verilog', 'VHDL', 'vcd', 'fst', 'development', 'hardware', 'rtl', 'simulation', 'verification', 'FPGA'],
    packages=find_packages(),
    python_requires='>=3.9, <4',
    install_requires=['lark'],
    extras_require={
        'dev': ['ruff'],
        'test': ['coverage'],
    },
    package_data={
        'wal': ['libs/std/std.wal', 'libs/std/std.wo', 'libs/std/module.wal', 'libs/std/module.wo']
    },

    entry_points={
        'console_scripts': [
            'wal=wal.wal:run',
            'walc=wal.walc:run',
            'wawk=wawk.wawk:run',
        ],
    },
)



'''The WAL core module'''

from wal.trace.container import TraceContainer
from wal.eval import SEval
from wal.reader import read_wal_sexpr, read_wal_sexprs, ParseError
from wal.ast_defs import Operator as Op
from wal.ast_defs import Symbol as S
from wal.ast_defs import UserOperator
from wal.ast_defs import WList
from wal.ast_defs import WalEvalError
from wal.passes import expand, optimize, resolve


class Wal:
    '''Main Wal class to be imported into other applications'''

    def __init__(self):  # pylint: disable=C0103
        self.traces = TraceContainer()
        self.eval_context = SEval(self.traces)
        self.eval_context.wal = self
        self.eval_context.eval(WList([Op.EVAL_FILE, S('std/std')]))
        self.eval_context.eval(WList([Op.EVAL_FILE, S('std/module')]))

    def load(self, file, tid='DEFAULT', from_string=False, keep_signals=None):
        '''Load trace from file and add it using id to WAL'''
        self.traces.load(file, tid, from_string=from_string, keep_signals=keep_signals)

    def step(self, steps=1, tid=None):
        '''Step one or all traces.
        If steps is not defined step trace(s) by +1.
        If no trace id (tid) is defined step all traces.
        '''
        return self.traces.step(steps, tid)

    def eval_str(self, txt, **args):
        '''Parses and evaluates the txt argument'''
        assert isinstance(txt, str)
        try:
            sexpr = read_wal_sexpr(txt)
            return self.eval(sexpr, **args)
        except ParseError as e:
            e.show()
            return None

    def eval(self, sexpr, **args):
        '''Evaluate the WAL expression sexpr and run passes'''
        # put passed arguments into context
        for name, val in args.items():
            if self.eval_context.global_environment.is_defined(name):
                self.eval_context.global_environment.write(name, val)
            else:
                self.eval_context.global_environment.define(name, val)

        res = None
        if sexpr:
            try:
                expanded = expand(self.eval_context, sexpr, parent=self.eval_context.global_environment)
                optimized = optimize(expanded)
                resolved = resolve(optimized, start=self.eval_context.global_environment.environment)
                res = self.eval_context.eval(resolved)
            except AssertionError as error:
                self.eval_context.print_error(sexpr, error)
                raise WalEvalError()

        # remove passed arguments from context
        for name, val in args.items():
            self.eval_context.global_environment.undefine(name)

        return res

    def run_str(self, txt, **args):
        '''Parses and runs the txt argument'''
        assert isinstance(txt, str)
        try:
            sexpr = read_wal_sexpr(txt)
            return self.run(sexpr, **args)
        except ParseError as e:
            e.show()
            return None

    def run(self, sexpr, **args):
        '''Evaluate the WAL expressions sexprs from Index 0'''
        res = None
        if sexpr:
            self.eval_context.reset()
            self.eval_context.eval(WList([Op.EVAL_FILE, S('std/std')]))
            self.eval_context.eval(WList([Op.EVAL_FILE, S('std/module')]))

            for name, val in args.items():
                self.eval_context.global_environment.define(name, val)

            expanded = expand(self.eval_context, sexpr, parent=self.eval_context.global_environment)
            optimized = optimize(expanded)
            resolved = resolve(optimized, start=self.eval_context.global_environment.environment)
            res = self.eval_context.eval(resolved)

        return res

    def run_file(self, filename):
        '''Executes a WAL program from a file'''
        with open(filename, 'r', encoding='utf-8') as fin:
            return self.eval(WList([Op.DO, *read_wal_sexprs(fin.read())]))


    def register_operator(self, name, function):
        self.eval_context.global_environment.define(name, UserOperator(name))
        self.eval_context.user_dispatch[name] = function


    def append_walpath(self, path):
        if path not in self.eval_context.walpath:
            self.eval_context.walpath.append(path)
            return True

        return False



'''Trace implementation for the CSV file format, as exported by Logic 2 '''
import re
import sys

from wal.trace.trace import Trace

class TraceCsv(Trace):
    '''Holds data for one csv trace.'''

    def __init__(self, filename, tid, container, from_string=False, keep_signals=None):
        super().__init__(tid, filename, container)
        self.timestamps = []
        self.lookup = None # performs index translation after set-sampling
        self.scopes = []
        self.rawsignals = []
        self.index2ts = []
        self.data = {}
        self.signalinfo = {}
        self.filename = filename
        self.keep_signals = set(keep_signals) if keep_signals else None
        if from_string:
            self.parse(filename)
        else:
            try:
                with open(filename) as f:
                    self.parse(f.read())
            except FileNotFoundError:
                print(f'Error while loading {filename}. File not found.')
                sys.exit(1)


        self.all_timestamps = self.timestamps.copy()
        self.index = 0
        self.max_index = len(self.index2ts) - 1
        self.signals = set(Trace.SPECIAL_SIGNALS + self.rawsignals)
        self.rawsignals_by_handle = [s for s in self.rawsignals]
        self.signals_by_handle = set(self.rawsignals_by_handle)

    def parse(self, csvdata):
        data = csvdata.strip().split("\n")  # assume row delimiter \n
        header = data[0].split(",") # assume col delimiter ,
        data = [line.split(",") for line in data[1:]]
        time_idx = header.index("Time [s]") # assume timestamp in seconds
        names = [v for v in header if v != "Time [s]"]
        self.data = {}
        
        for name in names:
            kind = "wire"
            width = 1
            orig_name = name
            # replace space with underscore
            name = re.sub(' ', '_', name)
            # remove slice info from names
            name = re.sub(r'\[[0-9]+:[0-9]+\]', '', name)
            # array signals should not clash with WAL operators
            name = re.sub(r'\[([0-9]+)\]', r'<\1>', name)
            name = re.sub(r'\(([0-9]+)\)', r'<\1>', name)
            header[header.index(orig_name)] = name
            self.data[name] = []
            if not self.keep_signals or (name in self.keep_signals):
                self.rawsignals.append(name)
                self.signalinfo[name] = {
                    'name': name,
                    'width': width,
                    'kind': kind,
                    'data': {}
                }
        

        pattern = re.compile(r"^(\d+)\.?(\d+)?$")
        for i in range(len(data)):
            # convert timestamp to nanoseconds
            m = pattern.match(data[i][time_idx])
            time_pre, time_post = m.groups("0")
            time_ns = int(f"{time_pre}{time_post.ljust(9, '0')}", base=10)
            
            for x in range(len(data[i])):
                if x != time_idx:
                    self.data[header[x]].append(data[i][x])
            self.timestamps.append(time_ns)
            self.index2ts.append(time_ns)

    def set_sampling_points(self, new_indices):
        '''Updates the indices at which data is sampled'''
        self.lookup = dict(enumerate(new_indices))
        new_timestamps = [self.all_timestamps[i] for i in new_indices]
        self.timestamps = list(dict.fromkeys(new_timestamps))
        self.timestamps = dict(enumerate(self.timestamps))
        # stores current time stamp
        self.index = 0
        self.max_index = len(self.timestamps.keys()) - 1

    def access_signal_data(self, name, index):
        value = None
        if self.lookup:
            value = self.data[name][self.lookup[index]]
        else:
            value = self.data[name][index]

        try:
            return int(value, 2)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return value

    def signal_width(self, name):
        '''Returns the width of a signal'''
        return self.signalinfo[name]['width']



from spade import Spade
from pathlib import Path
from wal.core import Wal
from wal.util import Colors, wal_str

class WalAnalysisPass:

    __wal_pass__ = True

    def __init__(self, pass_dirs, wavefile):
        self.pass_dirs = pass_dirs
        self.wavefile = wavefile
        self.testname = Path(wavefile).stem
        self.name = 'WAL Pass'
        wal = Wal()
        # add all the required dirs to walpath
        for pass_dir in pass_dirs:
            wal.append_walpath(pass_dir)

        wal.eval_str('(eval-file spade-definitions)')
        wal.eval_str(f'(define TEST "{self.testname}")')
        self.wal = wal
        self.config_file = wavefile.with_suffix('')
        wal.eval_str(f'(eval-file "{self.config_file}")')
        wal.load(str(wavefile))
        self.top = wal.eval_str('top_name')
        self.spade = Spade(self.top, 'build/state.ron')

        # register spade translation functions
        self.wal.register_operator('spade/translate', lambda seval, args: self.translate(seval, args).split('(')[0])
        self.wal.register_operator('spade/translate-full', lambda seval, args: self.translate(seval, args))
        self.wal.register_operator('log', lambda seval, args: self.log(seval, args))
        self.wal.register_operator('log/info', lambda seval, args: self.log(seval, args))
        self.wal.register_operator('log/analysis', lambda seval, args: self.log(seval, args, severity='analysis'))
        self.wal.register_operator('log/warn', lambda seval, args: self.log(seval, args, severity='warn'))
        self.wal.register_operator('log/error', lambda seval, args: self.log(seval, args, severity='error'))

    def run(self):
        pass

    def translate(self, seval, args):
        signal = seval.eval(args[0])
        if signal.startswith(self.top):
            # if the signal is abolute remove the top name
            signal = signal[len(self.top)+1:]

        value = seval.eval(args[1])
        if isinstance(value, int):
            value = f"{value:b}"

        return self.spade.translate_value(signal, value)

    def log(self, seval, args, severity='info'):
        def tostr(x):
            if isinstance(x, str):
                return x

            return wal_str(x)

        txt = ''.join(map(tostr, seval.eval_args(args)))
        if severity == 'info':
            print(f'[{Colors.green("INFO")}] {txt}')
        elif severity == 'analysis':
            print(f'[{Colors.blue("ANALYSIS")}] {txt}')
        elif severity == 'warn':
            print(f'[{Colors.yellow("WARNING")}] {txt}')
        elif severity == 'error':
            print(f'[{Colors.red("ERROR")}] {txt}')



'''Implementations for all list related functions'''
from wal.ast_defs import Operator, Symbol, Closure, WList

def op_list(seval, args):
    '''Constructs a new list filled with evaluated arguments'''
    return WList(seval.eval_args(args))


def op_first(seval, args):
    '''Returns the first element of list'''
    assert len(args) == 1, 'first: expects one argument (first list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), f'first: argument must be a list but is {type(evaluated)}'
    assert evaluated, 'first: argument must have length > 0'
    return evaluated[0]


def op_second(seval, args):
    '''Returns the second element of list'''
    assert len(args) == 1, 'second: expects one argument (second list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'second: argument must be a list'
    assert len(evaluated) > 1, 'second: argument must have length > 1'
    return evaluated[1]


def op_last(seval, args):
    '''Returns the last element of list'''
    assert len(args) == 1, 'last: expects one argument (last list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'last: argument must be a list'
    assert evaluated, 'last: argument must have length > 0'
    return evaluated[-1]


def op_rest(seval, args):
    '''Returns all elements but the first one from list'''
    assert len(args) == 1, 'rest: expects one argument (rest list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'rest: argument must be a list'
    return evaluated[1:] if len(evaluated) > 1 else []


def op_in(seval, args):
    '''Returns true if value is contained in list, array, or string'''
    assert len(args) >= 2, 'in: expects at least 2 arguments (in value* [list|array|str])'
    evaluated = seval.eval_args(args)
    res = True
    if isinstance(evaluated[-1], (WList, list)):
        for check in evaluated[:-1]:
            if check not in evaluated[-1]:
                res = False
                break
    elif isinstance(evaluated[-1], dict):
        key = '-'.join(map(lambda x: x.name if isinstance(x, Symbol) else str(x), evaluated[:-1]))
        res = key in evaluated[-1]
    elif isinstance(evaluated[-1], str):
        for check in evaluated[:-1]:
            assert isinstance(check, str), 'in: all arguments must be strings if the last argument is string'
            if check not in evaluated[-1]:
                res = False
                break
    else:
        assert False, 'in: expects list or array as last argument'
    return res


def op_map(seval, args):
    '''Applies function to each element of list. Functions must expect exactly one argument.'''
    assert len(args) == 2, 'map: expects two arguments (map function list)'

    arg = seval.eval(args[1])
    assert isinstance(arg, (WList, list)), 'map: second argument must be a list'
    res = []

    if isinstance(args[0], Operator):
        for element in arg:
            res.append(seval.eval(WList([args[0], WList([Operator.QUOTE, element])])))
    else:
        func = seval.eval(args[0])
        assert isinstance(func, Closure), 'map: first argument must be a function'
        for element in arg:
            res.append(seval.eval_closure(func, [[Operator.QUOTE, element]]))

    return res


def op_zip(seval, args):
    '''Combines two lists (a b c) (1 2 3) to ((a 1) (b 2) (c 3))'''
    assert len(args) == 2, 'zip: expects two arguments (zip list list)'
    evaluated = seval.eval_args(args)
    return list(map(list, zip(evaluated[0], evaluated[1])))


def op_max(seval, args):
    '''Returns the largest number from the list'''
    assert len(args) == 1, 'max: expects one argument (max list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'max: argument must be a list'
    return max(evaluated)


def op_min(seval, args):
    '''Returns the smalles number from the list'''
    assert len(args) == 1, 'min: expects one argument (min list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'min: argument must be a list'
    return min(evaluated)


def op_average(seval, args):
    '''Returns the average from the list'''
    assert len(args) == 1, 'average: expects one argument (min list)'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list)), 'average: argument must be a list'
    return sum(evaluated) / len(evaluated)


def op_length(seval, args):
    '''Returns the length of list'''
    assert len(args) == 1, 'length: expects one argument (length list), {args[0].line_info}'
    evaluated = seval.eval(args[0])
    assert isinstance(evaluated, (WList, list, str, dict)), 'length: argument must be a list'
    return len(evaluated)


def op_fold(seval, args):
    '''Performs a fold left on list with function f'''
    assert len(args) == 3, 'fold: expects 3 arguments (fold f acc data:list)'
    evaluated = seval.eval_args(args[1:])
    assert isinstance(evaluated[-1], (WList, list)), f'fold: last argument must be a list not {type(evaluated[-1])}'

    acc = evaluated[0]
    if isinstance(args[0], Operator):
        for element in evaluated[1]:
            acc = seval.eval(WList([args[0], WList([Operator.QUOTE, acc]), WList([Operator.QUOTE, element])]))
    else:
        func = seval.eval(args[0])
        assert isinstance(func, Closure), 'fold: first argument must be a function'
        for element in evaluated[1]:
            acc = seval.eval_closure(func, WList([WList([Operator.QUOTE, acc]), WList([Operator.QUOTE, element])]))

    return acc


def op_range(seval, args):
    '''Returns a list filled with numbers from arg0 to arg1. Arg1 can be omitted which
    results in a list 0..arg0'''
    assert len(args) >= 1 and len(args) <= 3, 'range: expects one or two arguments (range start:int end:int step:int)'
    evaluated = seval.eval_args(args)
    assert all(isinstance(e, int) for e in evaluated), 'range: all arguments must be ints'
    return list(range(*evaluated))


def op_for(seval, args):
    '''Loops over each element in seq, binds it to id and evaluates the body'''
    assert len(args) > 1, 'for: expects at least two arguments (for [id:symbol seq:list] body+)'
    assert isinstance(args[0], (WList, list)), 'for: first argument must be a tuple like [id:symbol seq:list]'
    assert len(args[0]) == 2,  'for: first argument must be a tuple like [id:symbol seq:list]'
    sym = args[0][0]
    sequence = seval.eval(args[0][1])
    assert isinstance(sequence, (WList, list, str)), 'for: seq in [id:symbol seq:list] must be a list'
    seval.environment.define(sym.name, 0)
    for value in sequence:
        seval.environment.write(sym.name, value)
        seval.eval_args(args[1:])


list_operators = {
    Operator.LIST.value: op_list,
    Operator.FIRST.value: op_first,
    Operator.SECOND.value: op_second,
    Operator.LAST.value: op_last,
    Operator.REST.value: op_rest,
    Operator.IN.value: op_in,
    Operator.MAP.value: op_map,
    Operator.MAX.value: op_max,
    Operator.MIN.value: op_min,
    Operator.AVERAGE.value: op_average,
    Operator.ZIP.value: op_zip,
    Operator.LENGTH.value: op_length,
    Operator.FOLD.value: op_fold,
    Operator.RANGE.value: op_range,
}



'''Implementations for all array related functions'''

from wal.ast_defs import Symbol, Operator, Closure


def op_array(seval, args):
    '''Creates a new array and populates it with the values passed in args'''
    res = {}
    for arg in args:
        assert len(arg) == 2, 'array: arguments must be (key:(int, str, Symbol) sexpr) tuples'
        key = seval.eval(arg[0])
        assert isinstance(key, (int, str, float, Symbol)), 'array: keys must be either int, string, float, or Symbol'

        if isinstance(key, Symbol):
            key = key.name
        elif isinstance(key, int):
            key = str(key)

        value = seval.eval(arg[1])
        res[key] = value
    return res


def to_str (sexpr):
    '''Converts argument sexpr to string. If sexpr is a symbol returns its name'''
    return sexpr.name if isinstance(sexpr, Symbol) else str(sexpr)


def op_seta(seval, args):
    '''Updates values depending on the tuples in args'''
    assert len(args) == 3, 'seta: requires three arguments. (seta array key:(int, str, symbol) value)'
    array = seval.eval(args[0])
    assert isinstance(array, dict), 'seta: must be applied on array'
    key = seval.eval(args[1])
    assert isinstance(key, (int, str, Symbol, float)), 'seta: key must be either int, string, float, or symbol'
    val = seval.eval(args[2])
    key = to_str(key)
    array[key] = val
    return array


def op_geta(seval, args):
    '''Returns the value at key from array.'''
    assert len(args) == 2, 'geta: requires at least three arguments. (geta array:(array, symbol) key:(int, str, symbol)'
    array = seval.eval(args[0])
    assert(isinstance(array, dict)), 'geta: first argument must be an array'
    key = seval.eval(args[1])
    assert isinstance(key, (int, str, float, Symbol)), 'geta: key must be either int, string, float or symbol'
    key = to_str(key)

    assert key in array, f'geta: key {key} not found'
    return array[key]


def op_dela(seval, args):
    '''Removes the value at key from array
    Returns the array without the removed value'''
    assert len(args) == 2, 'dela: requires at least two arguments. (dela array key)'
    array = seval.eval(args[0])
    assert(isinstance(array, dict)), 'dela: first argument must be an array'
    key = seval.eval(args[1])
    assert isinstance(key, (int, str, float, Symbol)), 'geta: key must be either int, string, float, or symbol'
    key = to_str(key)
    assert key in array, f'dela: key {key} not found'
    del array[key]
    return array


def op_mapa(seval, args):
    '''Applies function to every key value pair in array'''
    assert len(args) == 2, 'mapa: requires two arguments. (mapa function array)'
    func = seval.eval(args[0])
    assert isinstance(func, Closure), 'mapa: first argument must be a function'
    arg = seval.eval(args[1])
    assert isinstance(arg, dict), 'mapa: second argument must be an array'
    res = []
    for key, val in arg.items():
        # set current element
        res.append(seval.eval_closure(func, [[Operator.QUOTE, key], [Operator.QUOTE, val]]))
    return res


array_operators = {
    Operator.ARRAY.value: op_array,
    Operator.SETA.value: op_seta,
    Operator.GETA.value: op_geta,
    Operator.DELA.value: op_dela,
    Operator.MAPA.value: op_mapa
}
